#
# Scrobble Free - Scrobble to libre.fm and last.fm.
# Copyright (C) 2009-2010 Michael Gratton.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from __future__ import division

import cPickle as pickle
import os
import time
import urlparse

import gconf
import glib
import gnomekeyring
import gobject
import gtk
import rb
import rhythmdb

import scrobblefree.prefs as prefs
import scrobblefree.pylast as pylast


MAX_CONNECT_RETRY = 3
PLAY_DELAY = 10
SCROBBLE_DELAY = 240
HARD_FAILURE_LIMIT = 3
BACKOFF_DELAY = 60
BACKOFF_LIMIT = (120 * 60)

CLIENT_ID='rbx'
CLIENT_VER=str(rb.__version__)

#CLIENT_ID='tst'
#CLIENT_VER='1.0'

GCONF_KEYS = {
    'plugin': '/apps/rhythmbox/plugins/scrobblefree',
    'services': '/apps/rhythmbox/plugins/scrobblefree/services',
    'service': '/apps/rhythmbox/plugins/scrobblefree/service/%s',
    'type': '/apps/rhythmbox/plugins/scrobblefree/service/%s/type',
    'user': '/apps/rhythmbox/plugins/scrobblefree/service/%s/user',
    'uri': '/apps/rhythmbox/plugins/scrobblefree/service/%s/uri',
    }


class ScrobbleFreePlugin(rb.Plugin):

    def __init__(self):
        super(ScrobbleFreePlugin, self).__init__()
        self.shell = None
        self.gconf = None
        self.nm = None

        self.scrobblers = []
        self.current_track = None
        self.service_types = {
            'lastfm': LastFmService,
            'librefm': LibreFmService
            }

    def activate (self, shell):
        self.shell = shell
        shell_player = shell.get_player()
        self.player_cb_ids = (
            shell_player.connect('elapsed-changed',
                                 self._elapsed_changed_cb),
            shell_player.connect('playing-song-changed',
                                 self._playing_entry_changed_cb),
            )

        self.gconf = gconf.client_get_default()
        self.gconf.add_dir(GCONF_KEYS['plugin'],
                           gconf.CLIENT_PRELOAD_RECURSIVE)
        self.gconf_cb_ids = (
            self.gconf.notify_add(GCONF_KEYS['services'],
                                  self._load_services_cb),
            )
        self.gconf.notify(GCONF_KEYS['services'])

        if not self.scrobblers:
            # XXX this should probably be replaced by a schema,
            # but that's a PITA
            print 'creating default services'
            self.new_service('lastfm', 'lastfm')
            self.new_service('librefm', 'librefm')

        self.nm = NetworkManager(self._online_cb)

    @property
    def enabled_scrobblers(self):
        return [s for s in self.scrobblers if s.enabled]

    def deactivate(self, shell):
        self.nm.disconnect()

        for nid in self.gconf_cb_ids:
            self.gconf.notify_remove(nid)

        shell_player = self.shell.get_player()
        for sid in self.player_cb_ids:
            shell_player.disconnect(sid)

        for scrobbler in self.enabled_scrobblers:
            scrobbler.exit()

        self.current_track = None
        self.scobblers = []
        self.shell = None

    def is_configurable(self):
        return True

    def create_configure_dialog(self, dialog=None):
        if not dialog:
            controller = prefs.PrefsDialog(self.find_file('prefs.glade'),
                                           self.scrobblers)
            dialog = controller.dialog

        dialog.present()
        return dialog

    def new_service(self, svc_id, svc_type):
        self.gconf.set_string(GCONF_KEYS['type'] % svc_id, svc_type)
        services = self.gconf.get_list(GCONF_KEYS['services'],
                                       gconf.VALUE_STRING)
        services.append(svc_id)
        self.gconf.set_list(GCONF_KEYS['services'],
                            gconf.VALUE_STRING,
                            services)

    def _load_services_cb(self, client, cnxn_id, entry, data):
        services = []
        if (entry.value and
            entry.value.type == gconf.VALUE_LIST and
            entry.value.get_list_type() == gconf.VALUE_STRING):
            for value in entry.value.get_list():
                svc_id = value.get_string()
                svc_type = client.get_string(GCONF_KEYS['type'] % svc_id)

                service = self.service_types[svc_type](svc_id, client)
                service.connect('changed', self._service_changed_cb)
                glib.idle_add(self._process_connect, service)
                services.append(service)

        # XXX need to dispose of the old services here, disconnect signals, etc.
        self.scrobblers = services

    def _service_changed_cb(self, service):
        glib.idle_add(self._process_connect, service)

    def _process_connect(self, service):
        print 'Connecting.'
        retry_count = MAX_CONNECT_RETRY
        service.reset()
        while (retry_count > 0 and
               service.state in (ScrobblerService.NOT_CONNECTED,
                                 ScrobblerService.AUTH_ERROR)):
            service.handshake()
            retry_count -= 1
            if service.state == ScrobblerService.AUTH_ERROR:
                dialog = gtk.MessageDialog(
                    self.shell.props.window,
                    gtk.DIALOG_MODAL,
                    gtk.MESSAGE_QUESTION,
                    gtk.BUTTONS_OK,
                    'Please enter your password for:\n\n%s @ %s' %
                    (service.user, service.uri))
                entry = gtk.Entry()
                entry.props.visible = True
                entry.props.visibility = False
                dialog.vbox.pack_end(entry, True, False)
                if dialog.run():
                    service.setPassword(entry.props.text)
                dialog.destroy()
        if service.state == ScrobblerService.CONNECTED:
            service.flush_queue()
        return False # don't run again

    def _elapsed_changed_cb(self, player, elapsed):
        track = self.current_track
        if track:
            track.tick(elapsed)

            if track.play:
                track.played = True
                for scrobbler in self.enabled_scrobblers:
                    scrobbler.play(track)

    def _playing_entry_changed_cb(self, player, entry):
        # queue the current track for scrobbling, if needed
        track = self.current_track
        if track and track.scrobble:
            self.current_track = None
            for scrobbler in self.enabled_scrobblers:
                scrobbler.scrobble(track)
        elif track:
            print 'Not scrobbling track: %s' % track.title

        if entry:
            db = self.shell.props.db
            artist = db.entry_get(entry, rhythmdb.PROP_ARTIST)
            album = db.entry_get(entry, rhythmdb.PROP_ALBUM)
            title = db.entry_get(entry, rhythmdb.PROP_TITLE)
            duration = db.entry_get(entry, rhythmdb.PROP_DURATION)
            track_no = db.entry_get(entry, rhythmdb.PROP_TRACK_NUMBER)
            mbid = db.entry_get(entry, rhythmdb.PROP_MUSICBRAINZ_TRACKID)

            stream_title = \
                db.entry_request_extra_metadata(entry, 'rb:stream-song-title')
            if stream_title:
                title = stream_title
                if not artist:
                    artist = \
                        db.entry_request_extra_metadata(entry,
                                                        'rb:stream-song-artist')
                if not album:
                    album = \
                        db.entry_request_extra_metadata(entry,
                                                        'rb:stream-song-album')

            self.current_track = Track(artist, album, title,
                                       track_no, duration, mbid)

    def _online_cb(self):
        for scrobbler in self.enabled_scrobblers:
            scrobbler.online()
            scrobbler.flush_queue()


class ScrobblerService(gobject.GObject):


    NOT_CONNECTED = 0
    CONNECTED = 1
    AUTH_ERROR = 2
    CONFIG_ERROR = 3
    FATAL_ERROR = 4


    def __init__(self, svc_id, gconf_client, uri):
        self.__gobject_init__()

        self.id = svc_id
        self.gconf = gconf_client

        self.enabled = True
        self.state = ScrobblerService.NOT_CONNECTED
        self.uri = uri
        self.user = None
        self.pwd = None

        queue_path = os.path.join('scrobblefree', svc_id, 'queue')

        self.playing = None
        self.scrobble_queue = FileQueue(rb.find_user_data_file(queue_path))
        self._scrobbler = None

        self.hard_failures = 0
        self.next_retry = None
        self.backoff_delay = BACKOFF_DELAY

        gconf_client.notify_add(GCONF_KEYS['service'] % svc_id, self._gconf_cb)
        gconf_client.notify(GCONF_KEYS['service'] % svc_id)

    def handshake(self):
        if not self.uri or not self.user:
            self.state = ScrobblerService.CONFIG_ERROR
            print 'No uri/user for %s @ %s' % (self.user, self.uri)
            return

        if self.state == ScrobblerService.FATAL_ERROR:
            print 'Not connecting, a fatal error was reported.'
            return

        if (self.next_retry is not None and
            self.next_retry > time.time()):
            print 'Not connecting, backing off for a while.'
            return

        if self.pwd is None:
            parts = urlparse.urlparse(self.uri)
            try:
                pwds = gnomekeyring.find_network_password_sync(
                    self.user,
                    None,
                    parts.hostname,
                    None,
                    parts.scheme,
                    'password',
                    0)
            except gnomekeyring.Error,e:
                self.state = ScrobblerService.AUTH_ERROR
                print ('No password found for: %s, %s (%s)' %
                       (self.uri, self.user, e))
                return
            else:
                self.pwd = pylast.md5(pwds[0]['password'])

        self._scrobbler = pylast.Scrobbler(CLIENT_ID, CLIENT_VER,
                                           self.user, self.pwd, self.uri)

        try:
            self._scrobbler._do_handshake()
        except pylast.BadAuthentication, e:
            self.state = ScrobblerService.AUTH_ERROR
            print 'Authentication failed %s @ %s' % (self.user, self.uri)
        except pylast.BannedClient, e:
            self.state = ScrobblerService.FATAL_ERROR
            print 'Client is banned by service %s @ %s' % (self.user, self.uri)
        except pylast.BadTime, e:
            self.state = ScrobblerService.FATAL_ERROR
            print 'Check your system clock %s @ %s' % (self.user, self.uri)
        except Exception, e:
            self.state = ScrobblerService.NOT_CONNECTED
            print 'Handshake failed for %s @ %s: %s' % (self.user, self.uri, e)
        else:
            self.state = ScrobblerService.CONNECTED
            print 'Connected to %s' % self.uri

        if self.state == ScrobblerService.NOT_CONNECTED:
            self.hard_failures += 1
            self.backoff_delay *= 2
            if self.backoff_delay > BACKOFF_LIMIT:
                self.backoff_delay = BACKOFF_LIMIT
            self.next_retry = time.time() + self.backoff_delay
            print 'Backing off by %s' % self.backoff_delay
        else:
            self.hard_failures = 0
            self.backoff_delay = BACKOFF_DELAY
            self.next_retry = None

    def check_connection(self):
        if self.state == ScrobblerService.NOT_CONNECTED:
            self.handshake()
        return self.state == ScrobblerService.CONNECTED

    def connection_error(self, error):
        if isinstance(error, pylast.BadAuthentication):
            self.state = ScrobblerService.AUTH_ERROR
        elif isinstance(error, pylast.BadSession):
            self.state = ScrobblerService.NOT_CONNECTED
        else:
            self.hard_failures += 1

        if self.hard_failures == HARD_FAILURE_LIMIT:
            print "Too many hard failures, backing off."
            self.next_retry = time.time() + self.backoff_delay
            self.state = ScrobblerService.NOT_CONNECTED

    def online(self):
        self.hard_failures = 0
        self.backoff_delay = BACKOFF_DELAY
        self.next_retry = None

    def save(self):
        self.gconf.set_string(GCONF_KEYS['user'] % self.id, self.user)

    def reset(self):
        self.state = ScrobblerService.NOT_CONNECTED
        self.pwd = None

    def exit(self):
        self.reset()
        self.scrobble_queue.save()

    def setPassword(self, pwd):
        parts = urlparse.urlparse(self.uri)
        try:
            gnomekeyring.set_network_password_sync(None,
                                                   self.user,
                                                   None,
                                                   parts.hostname,
                                                   self.uri,
                                                   parts.scheme,
                                                   'password',
                                                   0,
                                                   pwd)
        except gnomekeyring.Error,e:
            print 'Unable to save password in keying: %s' % e
        else:
            self.pwd = pylast.md5(pwd)

    def flush_queue(self):
        if self.scrobble_queue:
            glib.idle_add(self._send_scrobble_queue_cb)

    def play(self, track):
        if self.state in (ScrobblerService.NOT_CONNECTED,
                          ScrobblerService.CONNECTED):
            glib.idle_add(self._send_playing_cb, track)

    def scrobble(self, track):
        if self.state in (ScrobblerService.NOT_CONNECTED,
                          ScrobblerService.CONNECTED):
            self.scrobble_queue.append(track)
            self.flush_queue()

    def _send_playing_cb(self, track):
        if self.check_connection():
            try:
                print "Now playing track: %s" % track.title
                self._scrobbler.report_now_playing(track.artist,
                                                   track.title,
                                                   track.album,
                                                   str(track.duration),
                                                   str(track.track),
                                                   str(track.mbid))
            except Exception, e:
                self.connection_error(e)
                print 'Error sending now playing: %s' % e
        else:
            print "Not sending now playing, not connected."
        return False

    def _send_scrobble_queue_cb(self):
        keep_going = False
        if self.check_connection():
            track = self.scrobble_queue.next()
            try:
                print "Scrobbling track: %s" % track.title
                self._scrobbler.scrobble(track.artist,
                                         track.title,
                                         track.started,
                                         pylast.SCROBBLE_SOURCE_USER,
                                         "",
                                         str(track.duration),
                                         track.album,
                                         str(track.track),
                                         str(track.mbid))
            except Exception, e:
                self.connection_error(e)
                print 'Error scrobbling: %s' % e
            else:
                track = self.scrobble_queue.pop()
                keep_going = len(self.scrobble_queue) > 0
        else:
            print "Not scrobling, not connected."
        return keep_going

    def _gconf_cb(self, client, cnxn_id, entry, data):
        self.user = self.gconf.get_string(GCONF_KEYS['user'] % self.id)
        uri = self.gconf.get_string(GCONF_KEYS['uri'] % self.id)
        if uri:
            self.uri = uri
        self.emit('changed')


gobject.type_register(ScrobblerService)
gobject.signal_new("changed", ScrobblerService,
                   gobject.SIGNAL_RUN_FIRST, gobject.TYPE_NONE, ())


class LastFmService(ScrobblerService):


        def __init__(self, svc_id, gconf_client):
            super(LastFmService, self).__init__(
                svc_id,
                gconf_client,
                'http://post.audioscrobbler.com/')

gobject.type_register(LastFmService)


class LibreFmService(ScrobblerService):


        def __init__(self, svc_id, gconf_client):
            super(LibreFmService, self).__init__(
                svc_id,
                gconf_client,
                'http://turtle.libre.fm/')

gobject.type_register(LibreFmService)


class Track(object):


    def __init__(self, artist, album, title, track, duration, mbid):
        self.started = int(time.time())
        self.play_duation = 0
        self.last_elapsed = 0
        self.played = False

        # ensure track data is sane-ish
        self.artist = artist or ''
        self.album = album or ''
        self.title = title or ''
        self.track = track or 0
        self.duration = duration or 0
        self.mbid = mbid or ''

    @property
    def play(self):
        return (not self.played and self.play_duation > PLAY_DELAY)

    @property
    def scrobble(self):
        # XXX also ensure wall clock elapsed time is reasonable
        return (self.play_duation > (self.duration // 2) or
                self.play_duation > SCROBBLE_DELAY)

    def tick(self, elapsed):
        # should be called around once every second while playing,
        # check if the difference is less than a small epsilon
        # however to allow some leeway
        delta = elapsed - self.last_elapsed
        if delta > 0 and delta <= 4:
            self.play_duation += delta
        self.last_elapsed = elapsed

    def __str__(self):
        return '<Track %s>' % self.__dict__


class NetworkManager(object):


    SIGNAL = 'StateChanged'
    NM = 'org.freedesktop.NetworkManager'


    def __init__(self, callback):
        self.callback = callback
        self.signal_match = None
        try:
            import dbus
            from dbus.mainloop.glib import DBusGMainLoop

            DBusGMainLoop(set_as_default=True)
            bus = dbus.SystemBus()
            self.signal_match = \
                bus.add_signal_receiver(self._state_changed_cb,
                                        "StateChanged",
                                        self.NM)
        except Exception, e:
            print "Error connecting to Network Manager: %s" % e

    def disconnect(self):
        if self.signal_match:
            self.signal_match.remove()
            self.signal_match = None

    def _state_changed_cb(self, state):
        if state == 3:
            print "Network is online"
            self.callback()


class FileQueue(object):


    def __init__(self, path):
        self.path = path
        self.queue = []

        print 'Loading queue: %s' % path
        try:
            dirname = os.path.dirname(path)
            if not os.path.isdir(dirname):
                os.makedirs(dirname)
            elif os.path.isfile(path):
                queue_file = open(self.path, 'rb')
                self.queue = pickle.load(queue_file)
                queue_file.close()
        except IOError, e:
            print "Error loading queue: %s" % e
        except EOFError:
            # empty file
            pass
        else:
            print 'Loaded %s tracks.' % len(self.queue)

    def __len__(self):
        return len(self.queue)

    def __iter__(self):
        return iter(self.queue)

    def save(self):
        print 'Saving queue: %s' % self.path
        try:
            queue_file = open(self.path, 'wb')
            pickle.dump(self.queue, queue_file, -1)
            queue_file.close()
        except IOError, e:
            print "Error saving queue: %s" % e
        else:
            print 'Queue saved.'

    def append(self, track):
        self.queue.append(track)

    def next(self):
        return self.queue[0]

    def pop(self):
        self.queue.pop(0)

